
#include <ctype.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <libgen.h>
#include <getopt.h>

#define SOH 0x01
#define STX 0x02
#define EOT 0x04
#define ACK 0x06
#define NAK 0x15
#define CAN 0x18

int MESSAGE_SIZE = 0;

typedef enum { UNKNOWN = 0, XMODEM = 1, YMODEM = 2 } xyproto_t;

char version[] = "1.0";
int debug = 1;
int do_unzip = 0;
const char optstring[] = "hv::z";
struct option opts[] = {
	{ "debug", optional_argument, NULL, 'v' },
	{ "unzip", no_argument, &do_unzip, 1 },
	{ "help", no_argument, NULL, 'h' },
	{ "dor-msg-size", required_argument, NULL, 1000 },
	{ NULL, 0, NULL, 0 }
};

int xymodem(const char*, int);
int xysendblk(int, unsigned char*, unsigned char);
unsigned short updcrc(register int, register unsigned int);
int dor_getch(int fd);
void hexdump(FILE*, void*, int);

void usage(void)
{
  fprintf(stderr,
	  "X/Y modem surface-to-DOM transfer program of K.Hanson\n"
	  "Version %s\n"
	  "usage : xymodem [options] <dor-device> <file> <flashname>\n"
	  "options :\n"
	  "\t--help or -h          show this message\n"
	  "\t--unzip or -z         unzip file upon delivery at DOM\n"
	  "\t--debug or -v [level] set debug level to [level]\n"
	  "Note that if file matches the pattern /.gz$/ then the unzip "
	  "option is implied.\n", version
	  );
}


static unsigned short crctab[256];

/*
 * updcrc macro derived from article Copyright (C) 1986 Stephen Satchell. 
 *  NOTE: First srgument must be in range 0 to 255.
 *        Second argument is referenced twice.
 * 
 * Programmers may incorporate any or all code into their programs, 
 * giving proper credit within the source. Publication of the 
 * source routines is permitted so long as proper credit is given 
 * to Stephen Satchell, Satchell Evaluations and Chuck Forsberg, 
 * Omen Technology.
 */
#ifndef SLOW_BROKEN_CRC_ALGORITHM
#define updcrc(cp, crc) ( crctab[((crc >> 8) & 255)] ^ (crc << 8) ^ cp)
#endif

#define max(x, y) ((x) < (y) ? (y) : (x))
#define min(x, y) ((x) < (y) ? (x) : (y))

int main(int argc, char **argv)
{
	int c;
	char *devicefile, *sendfile, *flashname;
	const char ymodem1k[] = "ymodem1k\r";
	char cmd[80];

	while ((c = getopt_long(argc, argv, optstring, opts, NULL)) != EOF) {
		switch (c) {
		case 'h':
			usage();
			exit(1);
		case 'z': /* Unzip at DOM */
			do_unzip = 1;
			break;
		case 'v':
			if (optarg)
				debug = atoi(optarg);
			else
				debug++;
			break;
		case 1000:
			MESSAGE_SIZE = atoi(optarg);
			break;
		}
	}
	
	if (argc - optind < 3) {
		usage();
		exit(1);
	}

	devicefile = argv[optind++];
	sendfile   = argv[optind++];
	flashname  = argv[optind++];

	/* 	if (!MESSAGE_SIZE) {
 		FILE *pf = fopen("/proc/driver/domhub/bufsiz", "r");
 		if (!pf) {
 			fprintf(stderr, "ERROR - procfile not found\n");
 			exit(1);
 		}
 		fscanf(pf, "%d", &MESSAGE_SIZE);
 		fclose(pf);
		}	*/
        if (!MESSAGE_SIZE) {
	  FILE *pf = fopen("/proc/driver/domhub/bufsiz", "r");
	  if (!pf) {
	    fprintf(stderr, "  procfile not found - setting (a guess) MESSAGE_SIZE to 500\n");
	    fprintf(stderr, "    note, setups prior to 2004 used a value of 400 \n");
	    MESSAGE_SIZE = 500;
	  } else {
	    fscanf(pf, "%d", &MESSAGE_SIZE);
	    fclose(pf);
	  }
        }

	/* Now have wiped out iceboot 2x by not unzipping */
	if (!strcmp(sendfile+strlen(sendfile)-3, ".gz")) do_unzip = 1;

	int devfd = open(devicefile, O_RDWR);
	if (devfd < 0) {
		perror(argv[0]);
		exit(1);
	}

	write(devfd, ymodem1k, strlen(ymodem1k));
	if (xymodem(sendfile, devfd)) return 1;

	/* wait for program prompt */
	while (dor_getch(devfd) != '>');
	
	/* Now program the flash - can take some time */
	if (do_unzip) {
		strcpy(cmd, "gunzip\r\n");
		write(devfd, cmd, strlen(cmd));
		while (dor_getch(devfd) != '>');
	}
	sprintf(cmd, "s\" %s\" create\r\n", flashname);
	write(devfd, cmd, strlen(cmd));
	while (dor_getch(devfd) != '>');
	close(devfd);

	return 0;
}

int xymodem(const char *filename, int rx)
{
	int bsiz = 0;
	unsigned char blknum = 0;
	unsigned char buf[1029];
	const unsigned char eot = EOT;
	unsigned char ack;
	xyproto_t protocol = UNKNOWN;
	int ifd, filesize;
	int cnak = 0;
	int handshake = 0;
	struct stat s;

	ifd = open(filename, O_RDONLY);
	if (ifd < 0) {
		perror("xymodem");
		return 1;
	}

	/* Stat the file to get its size */
	fstat(ifd, &s);
	filesize = (int) s.st_size;

	/* Wait on 1st NAK */
	while (!handshake) {
		int nak = dor_getch(rx);
		switch (nak) {
		case NAK: /* Normal XMODEM protocol */
			protocol = XMODEM;
			bsiz = 128;
			handshake = 1;
			break;
		case 'C': /* XMODEM/CRC or YMODEM enhancement */
			protocol = YMODEM;
			bsiz = 1024;
			handshake = 1;
			break;
		case EOF:
			if (++cnak > 10) handshake = -1;
			break;
		}
	}

	if (protocol == UNKNOWN) return 1;

	/* Handle YMODEM block */
	if (protocol == YMODEM) {
		char *base, *ptr;
		char fcpy[100];

		strncpy(fcpy, filename, 100);
		base = basename(fcpy);
		memset(buf, 0, sizeof(buf));
		
		buf[0] = SOH;
		ptr = strcpy(buf+3, base) + strlen(base) + 1;
		sprintf(ptr, "%d ", filesize);
		if (xysendblk(rx, buf, blknum++))
			return 1;
		/* Ymodem now sends a 'C' for something */
		ack = dor_getch(rx);
	}
	
	while (1) {

		memset(buf, 0, sizeof(buf));
		if (read(ifd, buf+3, bsiz) <= 0)
			break;

		switch (protocol) {
		case UNKNOWN:
			abort();
		case XMODEM:
			buf[0] = SOH;
			break;
		case YMODEM:
			buf[0] = STX;
			break;
		}
		if (xysendblk(rx, buf, blknum++)) 
			return 1;
	}
	
	/* Send the final break character */
	write(rx, &eot, 1);
	ack = dor_getch(rx);
	write(rx, &eot, 1);
	ack = dor_getch(rx);
	ack = dor_getch(rx);
	
	/* send a block /w/ null filename to stop transmission */
	memset(buf, 0, 133);
	buf[0] = SOH;
	xysendblk(rx, buf, 0);

	close(ifd);

	return 0;
}

int xysendblk(int fd, unsigned char *buf, unsigned char blknum)
{
	int i, blen, ack;
	unsigned short crc16;

	buf[1] = blknum;
	buf[2] = ~blknum;

	switch (buf[0]) {
	case SOH:
		blen = 133;
		break;
	case STX:
		blen = 1029;
		break;
	default:
		abort();
	}

	/* Calculate CRC-16 */
	for (crc16 = 0, i = 3; i < blen; i++)
		crc16 = updcrc(buf[i], crc16);
	buf[blen - 2] = (unsigned char) (crc16 >> 8);
	buf[blen - 1] = (unsigned char) (crc16 & 0xff);

	if (debug > 5) {
		/* Re-Calculate CRC-16 */
		for (crc16 = 0, i = 3; i < blen; i++)
			crc16 = updcrc(buf[i], crc16);
		fprintf(stderr, "CRC: %d\n", (int) crc16);
	}

	fprintf(stderr, "Sending block %d\n", blknum);
	if (debug > 4) hexdump(stderr, buf, blen);

	while (1) {
	        int pos = 0;
		while (pos < blen) {
			int nw = write(fd, buf+pos, min(blen-pos, MESSAGE_SIZE));
			if (nw < 0) {
				perror("xysendblk");
				exit(1);
			}
			pos += nw;
		}
		ack = dor_getch(fd);
		switch (ack) {
		case ACK:
			return 0;
		case NAK:
			break;
		case CAN:
			return 1;
		default:
			return 1;
		}
	}
}

int dor_getch(int fd)
{
	unsigned char buf[MESSAGE_SIZE];
	fd_set rs;
	struct timeval timeout = { 1, 0 };
	FD_ZERO(&rs);
	FD_SET(fd, &rs);
	if (select(fd+1, &rs, NULL, NULL, &timeout)) {
		int nr = read(fd, buf, MESSAGE_SIZE);
		if (debug > 1) {
			fprintf(stderr, "Read %d char %2.2x (%c) from DOM\n", 
				nr, buf[0], isprint(buf[0]) ? buf[0] : '.');
			if (debug > 3 && nr > 1)
				hexdump(stderr, buf, nr);
		}
		return buf[0];
	} 
	fprintf(stderr, "dor_getch timed out\n");
	return EOF;
}

/* crctab calculated by Mark G. Mendel, Network Systems Corporation */
static unsigned short crctab[256] = {
    0x0000,  0x1021,  0x2042,  0x3063,  0x4084,  0x50a5,  0x60c6,  0x70e7,
    0x8108,  0x9129,  0xa14a,  0xb16b,  0xc18c,  0xd1ad,  0xe1ce,  0xf1ef,
    0x1231,  0x0210,  0x3273,  0x2252,  0x52b5,  0x4294,  0x72f7,  0x62d6,
    0x9339,  0x8318,  0xb37b,  0xa35a,  0xd3bd,  0xc39c,  0xf3ff,  0xe3de,
    0x2462,  0x3443,  0x0420,  0x1401,  0x64e6,  0x74c7,  0x44a4,  0x5485,
    0xa56a,  0xb54b,  0x8528,  0x9509,  0xe5ee,  0xf5cf,  0xc5ac,  0xd58d,
    0x3653,  0x2672,  0x1611,  0x0630,  0x76d7,  0x66f6,  0x5695,  0x46b4,
    0xb75b,  0xa77a,  0x9719,  0x8738,  0xf7df,  0xe7fe,  0xd79d,  0xc7bc,
    0x48c4,  0x58e5,  0x6886,  0x78a7,  0x0840,  0x1861,  0x2802,  0x3823,
    0xc9cc,  0xd9ed,  0xe98e,  0xf9af,  0x8948,  0x9969,  0xa90a,  0xb92b,
    0x5af5,  0x4ad4,  0x7ab7,  0x6a96,  0x1a71,  0x0a50,  0x3a33,  0x2a12,
    0xdbfd,  0xcbdc,  0xfbbf,  0xeb9e,  0x9b79,  0x8b58,  0xbb3b,  0xab1a,
    0x6ca6,  0x7c87,  0x4ce4,  0x5cc5,  0x2c22,  0x3c03,  0x0c60,  0x1c41,
    0xedae,  0xfd8f,  0xcdec,  0xddcd,  0xad2a,  0xbd0b,  0x8d68,  0x9d49,
    0x7e97,  0x6eb6,  0x5ed5,  0x4ef4,  0x3e13,  0x2e32,  0x1e51,  0x0e70,
    0xff9f,  0xefbe,  0xdfdd,  0xcffc,  0xbf1b,  0xaf3a,  0x9f59,  0x8f78,
    0x9188,  0x81a9,  0xb1ca,  0xa1eb,  0xd10c,  0xc12d,  0xf14e,  0xe16f,
    0x1080,  0x00a1,  0x30c2,  0x20e3,  0x5004,  0x4025,  0x7046,  0x6067,
    0x83b9,  0x9398,  0xa3fb,  0xb3da,  0xc33d,  0xd31c,  0xe37f,  0xf35e,
    0x02b1,  0x1290,  0x22f3,  0x32d2,  0x4235,  0x5214,  0x6277,  0x7256,
    0xb5ea,  0xa5cb,  0x95a8,  0x8589,  0xf56e,  0xe54f,  0xd52c,  0xc50d,
    0x34e2,  0x24c3,  0x14a0,  0x0481,  0x7466,  0x6447,  0x5424,  0x4405,
    0xa7db,  0xb7fa,  0x8799,  0x97b8,  0xe75f,  0xf77e,  0xc71d,  0xd73c,
    0x26d3,  0x36f2,  0x0691,  0x16b0,  0x6657,  0x7676,  0x4615,  0x5634,
    0xd94c,  0xc96d,  0xf90e,  0xe92f,  0x99c8,  0x89e9,  0xb98a,  0xa9ab,
    0x5844,  0x4865,  0x7806,  0x6827,  0x18c0,  0x08e1,  0x3882,  0x28a3,
    0xcb7d,  0xdb5c,  0xeb3f,  0xfb1e,  0x8bf9,  0x9bd8,  0xabbb,  0xbb9a,
    0x4a75,  0x5a54,  0x6a37,  0x7a16,  0x0af1,  0x1ad0,  0x2ab3,  0x3a92,
    0xfd2e,  0xed0f,  0xdd6c,  0xcd4d,  0xbdaa,  0xad8b,  0x9de8,  0x8dc9,
    0x7c26,  0x6c07,  0x5c64,  0x4c45,  0x3ca2,  0x2c83,  0x1ce0,  0x0cc1,
    0xef1f,  0xff3e,  0xcf5d,  0xdf7c,  0xaf9b,  0xbfba,  0x8fd9,  0x9ff8,
    0x6e17,  0x7e36,  0x4e55,  0x5e74,  0x2e93,  0x3eb2,  0x0ed1,  0x1ef0
};

